{{ ************************************
   *  Calendar Object                 *
   *  (C) 2007 Gary Noel Boone, PhD   *
   ************************************

  This module provides two public date related functions. These are intended to be used
  with the Parallax GPS module for localizing the GPS time and date. 

  1.    ConvertUTCtoLocaltime()

  Given a desired offset (Pacific Time is -8 hours), a UTC date/time, and whether to allow
  Daylight Saving Time, this function will localize a UTC time.

  Features:
        - accounts for leap years
        - automatically converts to Daylight Saving Time if necessary 
        - includes a flag to skip conversion to DST for states that don't use it, such as
        Arizona and Hawaii.
                
  Limitations:
        - requires Western (negative) time offsets. The algorithm saves some memory by only
        handling the decrementing of days, so cannot account for positive time changes. It would
        be easy to modify the routines if positive offsets are desired. 
        
        - only works for 2007 and beyond. This module is intended to be used with the GPS module,
        so does not need to consider past dates. The start and end dates of Daylight Saving Time
        changed in 2007. This module encodes the current rules, although it would be easy to add
        previous rules if desired.
    

   To use it, you need to provide two values in addition to the UTC date and time:
        - the desired time zone
        - whether or not to apply Daylight Saving Time adjustments. 
  
   U.S. Time Zone names and offsets from UTC: 
        Eastern Standard Time                 -5
        Central Standard Time                 -6
        Mountain Standard Time                -7
        Pacific Standard Time                 -8
        Alaskan Standard Time (Alaska)        -9
        Hawaii-Aleutian Standard Time        -10


  2.    DateOfNthMday()

  This function is used by ConvertUTCtoLocaltime() but may be useful on its own. It answers questions
  such as, "What is the third Saturday in August?" The function is used by ConvertUTCtoLocaltime() to
  determine the start (2nd Sunday in March) and the end (1st Sunday in November) of Daylight Saving
  Time.

  Features:
        - accounts for leap years
                
  Limitations:
        - works only for the years from 1700 through 2400




  References:

  The DateOfNthMday() function uses an algorithm found here:
  
        http://www.cs.umd.edu/~khennacy/research/cell/calendar.pdf

  Here's a useful site for testing date/times

        http://www.timeanddate.com/worldclock/converter.html

  It's easy to plug in examples directly into the get variables in the URL
        http://www.timeanddate.com/worldclock/converted.html?month=3&day=12&year=2007&hour=6&min=35&sec=0&p1=0&p2=224


  
}}




CON
  '' These constants allow convenient calling syntax for testing functions that require zero offset arguments.  
  ' days, zero offset
  #0, Sun, Mon, Tue, Wed, Thu, Fri, Sat 
  ' months, zero offset
  #0, Jan, Feb, Mar, Apr, May, Jun, Jul, Aug, Sep, Oct, Nov, Dec

VAR
  long  _year             ' year as four digit
  long  _mon              ' month, zero offset
  long  _day              ' day, zero offset
  long  _hours            ' hours, zero offset, 24 hour day
  long  _mins             ' minutes
  long  _secs             ' seconds

      
DAT
              
        '' tables for conversion
        Cal_M           long    6, 2, 2, 5, 0, 3, 5, 1, 4, 6, 2, 4
        Cal_C           long    5, 3, 1, 0, 5, 3, 1, 0               ' valid from 1700 through 2400       
        DaysInMonth     long    31,28,31,30,31,30,31,31,30,31,30,31

        '' storage for conversion to strings
        _date_str       byte    "mm/dd/yy",0
        _time_str       byte    "hh:mm:ss",0
                        
        

PUB DateOfNthMday(count, D, mm, yyyy) | M, Y, C, yy, dd, max_days
{{ Return the day number of the Nth day type of the given mon and year. For example, what day is
   is the 3rd Wednesday of January, 2004? The answer is January 21st. Calling

      Cal.DateOfNthMday(3, Cal#Wed, Cal#Jan, 2004)

   returns 21.

   Notes:
      - mm is the month counted from Jan=0
      - the day number returned is _not_ 0 offset. 1 indicates the first of the month.

   Algorithm:
      This function implements the formula:     dd = (W-1)*7 + [ D - M - Y ] mod 7
      as described in
          http://www.cs.umd.edu/~khennacy/research/cell/calendar.pdf
      
}}

  yy := yyyy // 100                                     ' 2-digit year
  C :=  Cal_C[yyyy/100-17]                              ' century
  Y := Remainder(C + yy + yy/4, 7) 
  M := Cal_M[mm]
  if (mm==Jan OR mm==Feb) AND ItsALeapYear(yyyy)
    M -= 1
  dd := (count-1)*7 + Remainder(D - M - Y, 7)

  ' check we didn't exceed num days in month
  max_days := DaysInMonth[mm]
  if mm==Feb AND ItsALeapYear(yyyy)
    max_days += 1
  if dd > max_days
    return -1
  return dd


PRI ItsALeapYear(year)
'' assume 4 digit year.
  return year//400==0 OR (year//100<>0 AND year//4==0)     ' leap year?


PRI  Remainder(x,y)
'' Return x // y, assuming rounding toward -infinity, instead of the rounding toward
'' 0 assumed by the current implementation of Spin. Use whenever x can be negative.
  if x < 0
    return x // y + y
  return x // y         

  
PUB ConvertUTCtoLocaltime(offset, yyyy, mo, dd, hh, mm, ss, allowDST, localDateStr, localTimeStr)
{{ Given:
    - a timezone offset from UTC in hours,
    - a date and time in UTC,
    - a flag to allow DST,
    - and two strings for results
   return the localized date and time formatted as mm/dd/yy and hh/mm/ss
                                
  Notes:  
    - values passed in are not zero offset. Jan 1 is passed as 1 and 1
    - the year is 4 digit
    - localDateStr and localTimeStr are assumed to be addresses (ie, passed in with the @ operator).

  Example:
    OBJ
      Term  : "tv_terminal"
      Cal   : "Calendar"

    PUB PrintDateAndTime | dateStr, timeStr
      ' convert June 2, 2007, 01:34:10 to Pacific time, allowing DST. Return the resulting strings
      ' in dateStr and timeStr 
      Cal.ConvertUTCtoLocaltime(Cal#Pacific, 2007, 6, 2, 1, 34, 10, TRUE, @dateStr, @timeStr)
      Term.str(STRING("Date:"))
      Term.str(dateStr)
      Term.out(13)                                        ' carriage return
      Term.str(STRING("Time:"))
      Term.str(timeStr)
      Term.out(13)                                        ' carriage return

  Algorithm:
  The algorithm is designed to only substract hours. As a result, dates can only go backwards, so
  only a decrement date function is required.
      1. Calculate localtime using offset+1, ie, a DST offset
      2. Check DST start/end. If in DST, done adjusting time
      3. Not in DST? Subtract 1 hour, readjust the date
      4. Write the time and date strings. Return them.
  Because the initially localized time is in DST, the check for the start of DST will compare 3am
  to the local time instead of 2am.
}}

  LoadVars(yyyy, mo, dd, hh, mm, ss)

  if allowDST                                           ' don't apply DST offset if DST disabled
    offset += 1                                         ' set to DST offset
    ApplyNegativeOffset(||offset)                       ' pass absolute offset
    if !ItsDaylightSavingTime                           ' if not DST,
      ApplyNegativeOffset(1)                            ' go back one more hour 
  else
    ApplyNegativeOffset(||offset) 
    
  LONG[localDateStr] := WriteDate                        ' write and return string
  LONG[localTimeStr] := WriteTime                       


PRI LoadVars(yyyy, mo, dd, hh, mm, ss)
'' load the Calendar module's variables, converting days and months to zero offsets. (eg, Jan=0 instead of 1)
  _year   := yyyy
  _mon    := mo-1               ' make zero offset
  _day    := dd-1               ' make zero offset
  _hours  := hh
  _mins   := mm
  _secs   := ss


PRI ApplyNegativeOffset(offset)
  '' Apply the given offset to the date stored in the VARs
  '' Adjust to previous date, if necessary.   
  _hours -= offset
  if _hours < 0
    _hours += 24
    SetPreviousDay

                                                  
PRI SetPreviousDay
  '' assuming a date is stored in the vars, roll the date back one day
  _day -= 1
  if _day < 0                                           ' need previous month
    _mon -= 1
    if _mon < 0                                         ' need previous year
      _year -= 1
      _mon := Dec                                       ' 0 indexed
    _day := DaysInMonth[_mon]-1                         ' _mon is 0 indexed, make result 0 indexed
    if _mon==Feb AND ItsALeapYear(_year)
      _day := 28                                        ' 29th, but stored w/ 0 offset        

      
PRI ItsDaylightSavingTime | changeDay
{{ Return true if the date stored in the module vars is in the Daylight Saving Time period.
   The function looks at the date, quickly returning if the month is all DST or all non-DST.
   Then it looks at the days, again returning if the day is all DST or non-DST.
   It only has to check the time on DST change days.
   For the days on which DST changes, this function assumes that the time stored in the module
   vars is already in DST. So the standard 2am is already 3am.
  
   The Daylight Saving Time rules changed for 2007. This function uses the 2007 DST start and
   end rules and is therefore valid only for 2007 and beyond. 
  
   2007 DST Rules:
   Daylight Saving Time begins at 2:00 a.m. local time on the second Sunday in March.
   Daylight Saving Time ends at 2:00 a.m. local time on the first Sunday in November.
}}
  ' check months
  if _mon<Mar OR _mon>Nov
    return FALSE  
  if _mon>Mar AND _mon<Nov
    return TRUE

  ' in the month that DST starts, so check the day
  if _mon==Mar
    changeDay := DateOfNthMday(2, Sun, Mar, _year)-1    'change returned date to 0 offset
    if _day < changeDay
      return FALSE 
    if _day > changeDay
      return TRUE 
    '' in March change day, so check the hour
    return _hours => 3                                  'assume _hours is in DST, so 2am standard is 3am DST      

  ' in the month that DST ends, so check the day
  if _mon==Nov 
    changeDay := DateOfNthMday(1, Sun, Nov, _year)-1    'change returned date to 0 offset    
    if _day < changeDay
      return TRUE 
    if _day > changeDay
      return FALSE 
    '' in November change day, so check the hour
    return _hours < 2                                          
  

  
PRI WriteTime
'' Write the date in the module vars into an hh/mm/ss formatted string    
  _time_str[0] := "0" + _hours / 10   
  _time_str[1] := "0" + _hours // 10  
  _time_str[2] := ":"   
  _time_str[3] := "0" + _mins / 10  
  _time_str[4] := "0" + _mins // 10  
  _time_str[5] := ":"   
  _time_str[6] := "0" + _secs / 10  
  _time_str[7] := "0" + _secs // 10  
  _time_str[8] := 0   
  return @_time_str

PRI WriteDate | yy
'' Write the time in the module vars into a mm/dd/yy formatted string.
  _date_str[0] := "0" + (_mon+1) / 10                     ' convert from 0 offset   
  _date_str[1] := "0" + (_mon+1) // 10  
  _date_str[2] := "/"   
  _date_str[3] := "0" + (_day+1) / 10                     ' convert from 0 offset
  _date_str[4] := "0" + (_day+1) // 10   
  _date_str[5] := "/"
  yy := _year // 100   
  _date_str[6] := "0" + yy / 10   
  _date_str[7] := "0" + yy // 10   
  _date_str[8] := 0   
  return @_date_str

  